package com.tzld.longarticle.recommend.server.service.recommend.filter.strategy;

import com.ctrip.framework.apollo.spring.annotation.ApolloJsonValue;
import com.google.common.cache.Cache;
import com.google.common.cache.CacheBuilder;
import com.google.common.collect.Lists;
import com.google.common.util.concurrent.ThreadFactoryBuilder;
import com.tzld.longarticle.recommend.server.common.CommonThreadPoolExecutor;
import com.tzld.longarticle.recommend.server.common.enums.StatusEnum;
import com.tzld.longarticle.recommend.server.model.dto.Content;
import com.tzld.longarticle.recommend.server.model.entity.crawler.ArticleSensitive;
import com.tzld.longarticle.recommend.server.model.entity.longArticle.ArticleUnsafeTitle;
import com.tzld.longarticle.recommend.server.repository.crawler.ArticleSensitiveRepository;
import com.tzld.longarticle.recommend.server.repository.longArticle.ArticleUnsafeTitleRepository;
import com.tzld.longarticle.recommend.server.service.recommend.filter.FilterParam;
import com.tzld.longarticle.recommend.server.service.recommend.filter.FilterResult;
import com.tzld.longarticle.recommend.server.service.recommend.filter.FilterStrategy;
import com.tzld.longarticle.recommend.server.util.TitleSimilarCheckUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.collections4.CollectionUtils;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

import java.util.*;
import java.util.concurrent.*;
import java.util.stream.Collectors;

@Component
@Slf4j
public class SensitiveStrategy implements FilterStrategy {

    @Autowired
    private ArticleSensitiveRepository articleSensitiveRepository;
    @Autowired
    private ArticleUnsafeTitleRepository articleUnsafeTitleRepository;

    @ApolloJsonValue("${UnSafeTitles:[]}")
    private static List<String> UnSafeTitles;

    private final static ExecutorService pool = new CommonThreadPoolExecutor(
            5,
            5,
            0L, TimeUnit.SECONDS,
            new LinkedBlockingQueue<>(100),
            new ThreadFactoryBuilder().setNameFormat("SensitiveStrategy-%d").build(),
            new ThreadPoolExecutor.AbortPolicy());

    Cache<String, Boolean> similarityCache = CacheBuilder.newBuilder()
            .expireAfterWrite(60 * 60, TimeUnit.SECONDS).build();

    @Override
    public FilterResult filter(FilterParam param) {
        FilterResult filterResult = new FilterResult();
        List<String> result = Collections.synchronizedList(new ArrayList<>(param.getContents().size()));
        List<Content> filterContents = Collections.synchronizedList(new ArrayList<>());

//        Map<String, String> titleMd5Map = new HashMap<>();
//        Map<String, ArticleSensitive> articleSensitiveMap = new HashMap<>();
//        List<String> md5List = new ArrayList<>();
//        for (Content content : param.getContents()) {
//            String md5 = Md5Util.encoderByMd5(content.getTitle());
//            md5List.add(md5);
//            titleMd5Map.put(content.getTitle(), md5);
//        }
//        List<ArticleSensitive> articleSensitiveList = getArticleSensitive(md5List);
//        if (CollectionUtil.isNotEmpty(articleSensitiveList)) {
//            articleSensitiveMap = articleSensitiveList.stream().collect(Collectors.toMap(ArticleSensitive::getMd5, o -> o));
//        }
        List<String> allUnSafeTitles = UnSafeTitles;
        long t1 = System.currentTimeMillis();
        List<ArticleUnsafeTitle> articleUnsafeTitleList = articleUnsafeTitleRepository.getByStatus(StatusEnum.ONE.getCode());
        long t2 = System.currentTimeMillis();
        log.info("SensitiveStrategy get unsafe title DBList cost time:{}", t2 - t1);
        if (CollectionUtils.isNotEmpty(articleUnsafeTitleList)) {
            for (ArticleUnsafeTitle articleUnsafeTitle : articleUnsafeTitleList) {
                allUnSafeTitles.add(articleUnsafeTitle.getTitle());
            }
        }
        List<Set<Character>> unsafeTitleCache = TitleSimilarCheckUtil.makeCache(allUnSafeTitles);
        long t3 = System.currentTimeMillis();
        log.info("SensitiveStrategy make unsafe title cache cost time:{}", t3 - t2);
        List<String> allTitles = param.getContents().stream().map(Content::getTitle).collect(Collectors.toList());
        Map<String, Boolean> similarityMap = new HashMap<>(similarityCache.getAllPresent(allTitles));
        long t4 = System.currentTimeMillis();
        log.info("SensitiveStrategy get unsafe title cache cost time:{}", t4 - t3);
        List<List<Content>> partitions = Lists.partition(param.getContents(), 500);
        CountDownLatch cdl = new CountDownLatch(partitions.size());
        for (List<Content> partition : partitions) {
            pool.submit(() -> {
                try {
                    for (Content content : partition) {
                        try {
                            Boolean isSimilar;
                            if (similarityMap.containsKey(content.getTitle())) {
                                isSimilar = similarityMap.get(content.getTitle());
                            } else {
                                isSimilar = TitleSimilarCheckUtil.isDuplicateContentByCache(content.getTitle(), unsafeTitleCache,
                                        TitleSimilarCheckUtil.SIMILARITY_THRESHOLD);
                                similarityCache.put(content.getTitle(), isSimilar);
                            }
                            if (isSimilar) {
                                content.setFilterReason("安全违规");
                                filterContents.add(content);
                            } else {
                                result.add(content.getId());
                            }
                        } catch (Exception e) {
                            log.error("similar check error ", e);
                        }
                    }
                } finally {
                    cdl.countDown();
                }
            });
        }
        try {
            cdl.await();
        } catch (InterruptedException e) {
            log.error("similar check error ", e);
        }
        long t5 = System.currentTimeMillis();
        log.info("SensitiveStrategy filter cost time:{}", t5 - t4);
        filterResult.setContentIds(result);
        filterResult.setFilterContent(filterContents);
        return filterResult;
    }

    public List<ArticleSensitive> getArticleSensitive(List<String> md5List) {
        return articleSensitiveRepository.findAllById(md5List);
    }

}
